from __future__ import annotations

from datetime import datetime, timezone
from pathlib import Path
from typing import Union

from pydantic import AfterValidator
from typing_extensions import Annotated, TypeAlias, TypeVar

T = TypeVar("T")

StrPath: TypeAlias = Union[str, Path]
"""String or pathlib.Path"""

UTCDateTime = Annotated[datetime, AfterValidator(lambda dt: dt.astimezone(timezone.utc))]
"""datetime that's always in UTC."""

CollectionOf: TypeAlias = Union[list[T], tuple[T, ...], set[T]]
"""Type alias representing a union of list, tuple, and set."""



from __future__ import annotations

from functools import lru_cache
from typing import Callable, cast

from typing_extensions import ParamSpec, TypeVar

from nzb._types import CollectionOf

T = TypeVar("T")
P = ParamSpec("P")


def cache(func: Callable[P, T], /) -> Callable[P, T]:
    """Identical to functools.cache but with typing"""
    return cast(Callable[P, T], lru_cache(maxsize=None)(func))


def construct_meta_fields(
    title: str | None = None,
    passwords: CollectionOf[str] | str | None = None,
    tags: CollectionOf[str] | str | None = None,
    category: str | None = None,
) -> list[dict[str, str]]:
    """
    Constructor that constructs valid `<meta> .. </meta>` fields
    """

    meta = []

    if title:
        meta.append({"@type": "title", "#text": title})

    if passwords:
        if isinstance(passwords, str):
            meta.append({"@type": "password", "#text": passwords})
        else:
            for password in passwords:
                meta.append({"@type": "password", "#text": password})

    if tags:
        if isinstance(tags, str):
            meta.append({"@type": "tag", "#text": tags})
        else:
            for tag in tags:
                meta.append({"@type": "tag", "#text": tag})

    if category:
        meta.append({"@type": "category", "#text": category})

    return meta



from __future__ import annotations

from importlib import metadata

from typing_extensions import NamedTuple


class Version(NamedTuple):
    """Version tuple based on SemVer"""

    major: int
    """Major version number"""
    minor: int
    """Minor version number"""
    micro: int
    """Micro version number"""


def _get_version() -> str:
    """
    Get the version of nzb
    """
    try:
        return metadata.version("nzb")

    except metadata.PackageNotFoundError:
        return "0.0.0"



from __future__ import annotations

from collections import OrderedDict
from pathlib import Path
from xml.parsers.expat import ExpatError

from pydantic import validate_call
from typing_extensions import Literal, Self, overload
from xmltodict import parse as xmltodict_parse
from xmltodict import unparse as xmltodict_unparse

from nzb._exceptions import InvalidNZBError
from nzb._models import NZB
from nzb._parser import parse_doctype, parse_files, parse_metadata
from nzb._types import CollectionOf, StrPath
from nzb._utils import construct_meta_fields


class NZBParser:
    @validate_call
    def __init__(self, nzb: str, encoding: str | None = "utf-8") -> None:
        """
        Initialize the NZBParser instance.

        Parameters
        ----------
        nzb : str
            NZB content as a string.
        encoding : str, optional
            Encoding of the NZB content, defaults to `utf-8`.

        Raises
        ------
        InvalidNZBError
            Raised if the input is not valid XML.
            However, being valid XML doesn't guarantee it's a correctly structured NZB.
        """
        self.__nzb = nzb
        self.__encoding = encoding
        try:
            self.__nzbdict = xmltodict_parse(self.__nzb, encoding=self.__encoding)
        except ExpatError as error:
            raise InvalidNZBError(error.args[0])

    def parse(self) -> NZB:
        """
        Parse the NZB.

        Returns
        -------
        NZB
            NZB object representing the parsed NZB file.

        Raises
        ------
        InvalidNZBError
            Raised if the input is not valid NZB.
        """
        metadata = parse_metadata(self.__nzbdict)
        files = parse_files(self.__nzbdict)

        return NZB(metadata=metadata, files=files)

    @classmethod
    @validate_call
    def from_file(cls, nzb: StrPath, encoding: str | None = "utf-8") -> Self:
        """
        Create an NZBParser instance from an NZB file path.

        Parameters
        ----------
        nzb : StrPath
            File path to the NZB.
        encoding : str, optional
            Encoding of the NZB, defaults to `utf-8`.

        Returns
        -------
        NZBParser
            An NZBParser instance initialized with the content of the specified NZB file.
        """
        nzb = Path(nzb).expanduser().resolve().read_text(encoding=encoding)
        return cls(nzb, encoding)


class NZBMetaEditor:
    @validate_call
    def __init__(self, nzb: str, encoding: str = "utf-8") -> None:
        """
        Initialize the NZBMetaEditor instance.

        Parameters
        ----------
        nzb : str
            NZB content as a string.
        encoding : str, optional
            Encoding of the NZB content, defaults to `utf-8`.

        Raises
        ------
        InvalidNZBError
            Raised if the input is not valid XML.
            However, being valid XML doesn't guarantee it's a correctly structured NZB.
        """
        self.__nzb = nzb
        self.__encoding = encoding
        try:
            self.__nzbdict = xmltodict_parse(self.__nzb, encoding=self.__encoding)
        except ExpatError as error:
            raise InvalidNZBError(error.args[0])

    def __get_meta(self) -> list[dict[str, str]] | dict[str, str] | None:
        """
        Retrieve current metadata from the NZB.

        Returns
        -------
        list[dict[str, str]] | dict[str, str] | None
            The metadata as a list of dictionaries, a single dictionary, or `None` if not found.
        """
        return self.__nzbdict.get("nzb", {}).get("head", {}).get("meta")  # type: ignore

    @validate_call
    def set(
        self,
        *,
        title: str | None = None,
        passwords: CollectionOf[str] | str | None = None,
        tags: CollectionOf[str] | str | None = None,
        category: str | None = None,
    ) -> Self:
        """
        Set metadata fields in the NZB.
        This will also remove all existing metadata fields.

        Parameters
        ----------
        title : str, optional
            The title metadata field.
        passwords : CollectionOf[str] | str, optional
            Password(s) for the NZB file.
        tags : CollectionOf[str] | str, optional
            Tag(s) associated with the NZB file.
        category : str, optional
            Category of the NZB file.

        Returns
        -------
        Self
            Returns itself.
        """

        if title is None and passwords is None and tags is None and category is None:
            return self

        nzb = OrderedDict(self.__nzbdict["nzb"])

        nzb["head"] = {}
        nzb["head"]["meta"] = construct_meta_fields(title=title, passwords=passwords, tags=tags, category=category)
        nzb.move_to_end("file")
        self.__nzbdict["nzb"] = nzb
        return self

    @validate_call
    def append(
        self,
        *,
        title: str | None = None,
        passwords: CollectionOf[str] | str | None = None,
        tags: CollectionOf[str] | str | None = None,
        category: str | None = None,
    ) -> Self:
        """
        Append metadata fields to the existing metadata in the NZB.

        Parameters
        ----------
        title : str, optional
            The title metadata field.
        passwords : CollectionOf[str] | str, optional
            Password(s) for the NZB file.
        tags : CollectionOf[str] | str, optional
            Tag(s) associated with the NZB file.
        category : str, optional
            Category of the NZB file.

        Returns
        -------
        Self
            Returns itself.
        """

        meta = self.__get_meta()

        if meta is None:
            self.set(title=title, passwords=passwords, tags=tags, category=category)

        elif isinstance(meta, dict):
            new_meta = [meta]
            new_meta.extend(construct_meta_fields(title=title, passwords=passwords, tags=tags, category=category))
            self.__nzbdict["nzb"]["head"]["meta"] = new_meta

        else:
            meta.extend(construct_meta_fields(title=title, passwords=passwords, tags=tags, category=category))
            self.__nzbdict["nzb"]["head"]["meta"] = meta

        return self

    @overload
    def remove(self, key: Literal["title", "password", "tag", "category"]) -> Self: ...

    @overload
    def remove(self, key: str) -> Self: ...

    @validate_call
    def remove(self, key: Literal["title", "password", "tag", "category"] | str) -> Self:
        """
        Remove a metadata field from the NZB.
        If the same field is present multiple times, this will remove them all.

        Parameters
        ----------
        key : Literal["title", "password", "tag", "category"] | str, optional
            The metadata field to remove.

        Returns
        -------
        Self
            Returns itself.
        """

        meta = self.__get_meta()

        if meta is None:
            return self

        elif isinstance(meta, dict):
            if key == meta["@type"]:
                return self.clear()
            else:
                return self
        else:
            new_meta = [row for row in meta if row["@type"] != key]
            self.__nzbdict["nzb"]["head"]["meta"] = new_meta
            return self

    def clear(self) -> Self:
        """
        Clear all metadata fields from the NZB.

        Returns
        -------
        Self
            Returns itself.
        """
        try:
            del self.__nzbdict["nzb"]["head"]
        except KeyError:  # pragma: no cover
            pass

        return self

    @validate_call
    def save(self, filename: StrPath | None = None, overwrite: bool = False) -> Path:
        """
        Save the edited NZB to a file.

        Parameters
        ----------
        filename : StrPath, optional
            Destination path for saving the NZB.
            If not provided, uses the original file path if available.
            This will also create the path if it doesn't exist already.
        overwrite : bool, optional
            Whether to overwrite the file if it exists, defaults to `False`.

        Returns
        -------
        Path
            The path to the saved file.

        Raises
        ------
        FileNotFoundError
            If no filename is specified and the original file path is unknown.
        FileExistsError
            If the file exists and overwrite is `False`.
        """

        filename_from_self: Path | None = getattr(self, "__nzb_file", None)

        if filename is None:
            if filename_from_self is None:
                raise FileNotFoundError("No filename specified!")
            else:
                if overwrite:
                    outfile = filename_from_self
                else:
                    raise FileExistsError(filename_from_self)
        else:
            outfile = Path(filename).resolve()

        outfile.parent.mkdir(parents=True, exist_ok=True)
        unparsed = xmltodict_unparse(self.__nzbdict, encoding=self.__encoding, pretty=True, indent="    ")

        if doctype := parse_doctype(self.__nzb):
            # see: https://github.com/martinblech/xmltodict/issues/351
            nzb = unparsed.splitlines(keepends=True)
            nzb.insert(1, f"{doctype}\n")
            with open(outfile, "w", encoding=self.__encoding) as f:
                f.writelines(nzb)
        else:
            outfile.write_text(unparsed, encoding=self.__encoding)

        return outfile

    @classmethod
    @validate_call
    def from_file(cls, nzb: StrPath, encoding: str = "utf-8") -> Self:
        """
        Create an NZBMetaEditor instance from an NZB file path.

        Parameters
        ----------
        nzb : StrPath
            File path to the NZB.
        encoding : str, optional
            Encoding of the NZB, defaults to `utf-8`.

        Returns
        -------
        Self
            Returns itself.
        """
        data = Path(nzb).expanduser().resolve().read_text(encoding=encoding)
        instance = cls(data, encoding)
        setattr(instance, "__nzb_file", nzb)
        return instance



from __future__ import annotations

import re

from nzb._utils import cache


@cache
def _is_par2(filename: str) -> bool:
    """
    Determine if a given file is likely a par2 file.

    Parameters
    ----------
    filename : str

    Returns
    -------
    bool
    """
    if not filename:
        return False
    else:
        parsed = re.search(r"\.par2$", filename, re.IGNORECASE)
        return True if parsed else False


@cache
def _is_rar(filename: str) -> bool:
    """
    Determine if a given file is likely a rar file.

    Parameters
    ----------
    filename : str

    Returns
    -------
    bool

    Notes
    -----
    The regex used here was copy-pasted from SABnzbd:
    https://github.com/sabnzbd/sabnzbd/blob/02b4a116dd4b46b2d2f33f7bbf249f2294458f2e/sabnzbd/nzbstuff.py#L107
    """
    if not filename:
        return False
    else:
        parsed = re.search(r"(\.rar|\.r\d\d|\.s\d\d|\.t\d\d|\.u\d\d|\.v\d\d)$", filename, re.IGNORECASE)
        return True if parsed else False


@cache
def _is_obfuscated(filestem: str) -> bool:  # pragma: no cover
    """
    Determine if a given file stem is likely obfuscated.

    Parameters
    ----------
    filestem : str
        The filestem (i.e., the name of the file without the extension)

    Returns
    -------
    bool
        True if the filestem is considered obfuscated, False otherwise.

    Notes
    -----
    This function was copy-pasted from SABnzbd:
    https://github.com/sabnzbd/sabnzbd/blob/297455cd35c71962d39a36b7f99622f905d2234e/sabnzbd/deobfuscate_filenames.py#L104
    """
    if not filestem:
        return True

    # First: the patterns that are certainly obfuscated:

    # ...blabla.H.264/b082fa0beaa644d3aa01045d5b8d0b36.mkv is certainly obfuscated
    if re.findall(r"^[a-f0-9]{32}$", filestem):
        # exactly 32 hex digits, so:
        return True

    # 0675e29e9abfd2.f7d069dab0b853283cc1b069a25f82.6547
    if re.findall(r"^[a-f0-9.]{40,}$", filestem):
        return True

    # "[BlaBla] something [More] something 5937bc5e32146e.bef89a622e4a23f07b0d3757ad5e8a.a02b264e [Brrr]"
    # So: square brackets plus 30+ hex digit
    if re.findall(r"[a-f0-9]{30}", filestem) and len(re.findall(r"\[\w+\]", filestem)) >= 2:
        return True

    # /some/thing/abc.xyz.a4c567edbcbf27.BLA is certainly obfuscated
    if re.findall(r"^abc\.xyz", filestem):
        # ... which we consider as obfuscated:
        return True

    # Then: patterns that are not obfuscated but typical, clear names:

    # these are signals for the obfuscation versus non-obfuscation
    decimals = sum(1 for c in filestem if c.isnumeric())
    upperchars = sum(1 for c in filestem if c.isupper())
    lowerchars = sum(1 for c in filestem if c.islower())
    spacesdots = sum(1 for c in filestem if c == " " or c == "." or c == "_")  # space-like symbols

    # Example: "Great Distro"
    if upperchars >= 2 and lowerchars >= 2 and spacesdots >= 1:
        return False

    # Example: "this is a download"
    if spacesdots >= 3:
        return False

    # Example: "Beast 2020"
    if (upperchars + lowerchars >= 4) and decimals >= 4 and spacesdots >= 1:
        return False

    # Example: "Catullus", starts with a capital, and most letters are lower case
    if filestem[0].isupper() and lowerchars > 2 and upperchars / lowerchars <= 0.25:
        return False

    # Finally: default to obfuscated:
    return True  # default is obfuscated



from __future__ import annotations

from nzb._core import NZBMetaEditor, NZBParser
from nzb._exceptions import InvalidNZBError, NZBException
from nzb._models import NZB, File, Metadata, Segment
from nzb._version import Version, _get_version

__version__ = _get_version()
__version_tuple__ = Version(*[int(i) for i in __version__.split(".")])

__all__ = ("NZBParser", "NZBMetaEditor", "Metadata", "File", "Segment", "NZB", "NZBException", "InvalidNZBError")
